#
# houdini_html_u_c.py
#
# Copyright (C) 2017-2022 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
r"""A cmark implementation file."""

import re

from ..constants import parser as md_parser
from ..generic import _strncmp
from .buffer_c import (
    _cmark_cmark_strbuf_grow,
    _cmark_cmark_strbuf_put,
    _cmark_cmark_strbuf_putc,
    _cmark_cmark_strbuf_puts,
)
from .buffer_h import _cmarkCmarkStrbuf
from .houdini_h import _cmark_HOUDINI_UNESCAPED_SIZE
from .utf8_c import _cmark_cmark_utf8proc_encode_char

# License F applies to this file except for non derivative code:
# in that case the license header at the top of the file applies.
# See docs/copyright_license.rst


# Recursive function of a binary search.
# 0.30.
def _cmark_S_lookup(i: int, low: int, hi: int, s: str, length: int) -> str:
    j: int
    cmp: int = _strncmp(
        s, md_parser['cmark']['re']['ENTITIES']['entities'][i]['entity'],
        length)
    #  if (cmp == 0 && cmark_entities[i].entity[len] == 0) {
    if cmp == 0 and length == len(
            md_parser['cmark']['re']['ENTITIES']['entities'][i]['entity']):
        return md_parser['cmark']['re']['ENTITIES']['entities'][i]['bytes']
    elif cmp == -1 and i > low:
        j = i - int((i - low) / 2)
        if j == i:
            j -= 1
        return _cmark_S_lookup(j, low, i - 1, s, length)
    elif cmp == 1 and i < hi:
        j = i + int((hi - i) / 2)
        if j == i:
            j += 1
        return _cmark_S_lookup(j, i + 1, hi, s, length)
    else:
        return None


# 0.30.
def _cmark_S_lookup_entity(s: str, length: int):
    return _cmark_S_lookup(
        int(md_parser['cmark']['re']['ENTITIES']['CMARK_NUM_ENTITIES'] / 2), 0,
        md_parser['cmark']['re']['ENTITIES']['CMARK_NUM_ENTITIES'] - 1, s,
        length)


# 0.30.
def _cmark_houdini_unescape_ent(ob: _cmarkCmarkStrbuf, src: str,
                                size: int) -> int:
    i: int = 0

    if size >= 3 and src[0] == '#':
        codepoint: int = 0
        num_digits: int = 0
        max_digits: int = 7

        if re.match(r'\d', src[1]):

            i = 1
            while i < size and re.match(r'\d', src[i]):
                codepoint = (codepoint * 10) + (ord(src[i]) - ord('0'))

                if codepoint >= 0x110000:
                    # Keep counting digits but
                    # avoid integer overflow.
                    codepoint = 0x110000

                i += 1

            num_digits = i - 1
            max_digits = 7

        elif src[1] == 'x' or src[1] == 'X':
            i = 2
            while i < size and re.match(r'[\dA-Fa-f]', src[i]):
                codepoint = (codepoint * 16) + ((ord(src[i]) | 32) % 39 - 9)

                if codepoint >= 0x110000:
                    # Keep counting digits but
                    # avoid integer overflow.
                    codepoint = 0x110000

                i += 1

            num_digits = i - 2
            max_digits = 6

        if (num_digits >= 1 and num_digits <= max_digits and i < size
                and src[i] == ';'):
            if (codepoint == 0 or (codepoint >= 0xD800 and codepoint < 0xE000)
                    or codepoint >= 0x110000):
                codepoint = 0xFFFD

            _cmark_cmark_utf8proc_encode_char(codepoint, ob)
            return i + 1

    else:
        if size > md_parser['cmark']['re']['ENTITIES'][
                'CMARK_ENTITY_MAX_LENGTH']:
            size = md_parser['cmark']['re']['ENTITIES'][
                'CMARK_ENTITY_MAX_LENGTH']

        for i in range(
                md_parser['cmark']['re']['ENTITIES']
            ['CMARK_ENTITY_MIN_LENGTH'], size):
            if src[i] == ' ':
                break

            if src[i] == ';':
                entity: str = _cmark_S_lookup_entity(src, i)

                if entity is not None:
                    _cmark_cmark_strbuf_puts(ob, entity)
                    return i + 1

                break

    return 0


# 0.30.
def _cmark_houdini_unescape_html(
    ob: _cmarkCmarkStrbuf,
    src: str,
    size: int,
) -> int:
    i: int = 0
    org: int
    ent: int

    while i < size:
        org = i
        while i < size and src[i] != '&':
            i += 1

        if i > org:
            if org == 0:
                if i >= size:
                    return 0

                _cmark_cmark_strbuf_grow(ob,
                                         _cmark_HOUDINI_UNESCAPED_SIZE(size))

            _cmark_cmark_strbuf_put(ob, src[org:], i - org)

        # escaping
        if i >= size:
            break

        i += 1

        ent = _cmark_houdini_unescape_ent(ob, src[i:], size - i)
        i += ent

        # not really an entity
        if ent == 0:
            _cmark_cmark_strbuf_putc(ob, ord('&'))

    return 1


def _cmark_houdini_unescape_html_f(
    ob: _cmarkCmarkStrbuf,
    src: str,
    size: int,
):
    if not _cmark_houdini_unescape_html(ob, src, size):
        _cmark_cmark_strbuf_put(ob, src, size)



#!/usr/bin/env python3
#
# Copyright (C) 2022 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
r"""A simple benchmark file to be used to check for hidden errors as well."""

import csv
import ctypes
import multiprocessing
import platform
import random
import secrets
import string
import subprocess
import tempfile
import time
import traceback

import md_toc

# ~50M characters
CHAR_SIZE = 1024 * 1024 * 50
ITERATIONS = 100

# At what nTH character to put a header.
# MIN_HEADER_STEP must be >= 10.
MIN_HEADER_STEP = 10
MAX_HEADER_STEP = 50

# Do not use '#' as content to avoid triggering an empty link label exception.
alphabet: str = string.printable.replace('#', '')


def _generate_batch(size: int) -> bytes:
    return bytes(
        [secrets.choice(alphabet).encode('UTF-8')[0] for _ in range(size)])


def _generate_random_characters(size: int,
                                min_header_step: int = 10,
                                max_header_step: int = 1000) -> bytes:
    if min_header_step < 10 or max_header_step < min_header_step:
        # 10 - 1 = header (maximum 6) + '\n' + space + alphanum char
        raise ValueError

    print('generating random file...')

    secret_gen: random.SystemRandom = secrets.SystemRandom()
    alphanumerics: str = ''.join([string.ascii_letters, string.digits])

    # Batch size at 1% of total size.
    batch_size: int = int(CHAR_SIZE * 0.01)
    # Chunk size at 1 per 10**4 of the batch size.
    chunk_size: int = int(batch_size * 0.00001)

    with multiprocessing.Pool() as pool:
        results = pool.map(_generate_batch,
                           [batch_size] * (size // batch_size), chunk_size)

    string_buf = ctypes.create_string_buffer(
        b''.join(results),
        size=size + 1,
    )

    print('adding headers to file...')

    i: int = 0
    j: int = 1
    while i + j + 3 < size:
        newline = b'\n'
        heading = b'#' * j
        space = b' '
        random_char = secrets.choice(alphanumerics).encode('utf-8')

        ctypes.memmove(ctypes.byref(string_buf, i), newline, 1)
        ctypes.memmove(ctypes.byref(string_buf, i + 1), heading, len(heading))
        ctypes.memmove(ctypes.byref(string_buf, i + 1 + j), space, 1)

        # Replace next character so we are sure never to raise the empty
        # link label exception.
        ctypes.memmove(ctypes.byref(string_buf, i + 2 + j), random_char, 1)

        # Reset header level.
        j = (j % 6) + 1
        i += secret_gen.randrange(min_header_step, max_header_step)

    return string_buf.value


if __name__ == '__main__':
    ok: bool = True
    i: int = 0
    j: int = 0
    total_iterations: int
    percent_progress: int
    current_parser_counter: int = 0
    avg: list
    average: dict = dict()
    total: dict = dict()

    parsers = ['github', 'cmark', 'redcarpet', 'gitlab']
    parsers.sort()

    total_iterations = len(parsers) * ITERATIONS
    for current_parser_counter, p in enumerate(parsers):
        print('parser: ' + p + ', ' + str(ITERATIONS) + ' iterations with ' +
              str(CHAR_SIZE) + ' characters each')
        print('min step: ' + str(MIN_HEADER_STEP) + ' , max step: ' +
              str(MAX_HEADER_STEP))
        i = 0
        avg = list()
        while ok and i < ITERATIONS:
            print('parser ' + str(p) + ' (' + str(current_parser_counter + 1) +
                  ' of ' + str(len(parsers)) + '), iteration: ' + str(i + 1) +
                  ' of ' + str(ITERATIONS))
            percent_progress = (j / total_iterations) * 100
            print('total progress percent = ' + str(percent_progress))
            with tempfile.NamedTemporaryFile() as fp:
                fp.write(
                    _generate_random_characters(
                        CHAR_SIZE,
                        min_header_step=MIN_HEADER_STEP,
                        max_header_step=MAX_HEADER_STEP))
                print('building TOC...')
                try:
                    start = time.time()
                    md_toc.api.build_toc(filename=fp.name,
                                         parser=p,
                                         keep_header_levels=3)
                    end = time.time()
                    avg.append(end - start)
                    print('total_time: ' + str(avg[-1]))
                except Exception:
                    ok = False
                    traceback.print_exc()
            i += 1
            j += 1
        total[p] = sum(avg)
        average[p] = total[p] / len(avg)
        print('total = ' + str(total[p]) + '\n')
        print('avg = ' + str(average[p]) + '\n')

    # Write CSV file.
    # Fields
    #   0   md_toc git hash version
    #   1   parser name
    #   2   number of characters
    #   3   number of iterations
    #   4   min_header_step
    #   5   max_header_step
    #   6   total execution time in seconds
    #   7   average execution time in seconds
    #   8->n system information
    # File header:
    # md_toc_git_hash,markdown_parser,total_characters,iterations,min_header_step,max_header_step,total,avg,platform_python_version,platform_architecture,platform_machine,platform_python_implementation,platform_python_compiler,platform_libc_ver
    md_toc_version = subprocess.check_output(
        ['/usr/bin/git', 'rev-parse', 'HEAD']).decode('UTF-8').strip()
    with open('benchmark.csv', 'a') as csvfile:
        spamwriter = csv.writer(csvfile,
                                delimiter=',',
                                quotechar='|',
                                quoting=csv.QUOTE_MINIMAL)
        for p in parsers:
            spamwriter.writerow([
                md_toc_version, p, CHAR_SIZE, ITERATIONS, MIN_HEADER_STEP,
                MIN_HEADER_STEP, total[p], average[p],
                platform.python_version(), ' '.join(platform.architecture()),
                platform.machine(),
                platform.python_implementation(),
                platform.python_compiler(), ' '.join(platform.libc_ver())
            ])



#
# utf8_c.py
#
# Copyright (C) 2017-2022 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
r"""A cmark implementation file."""

import unicodedata

from ..constants import parser as md_parser
from .buffer_c import _cmark_cmark_strbuf_put
from .buffer_h import _cmarkCmarkStrbuf
from .cmark_ctype_c import _cmark_cmark_ispunct

# License D applies to this file except for non derivative code:
# in that case the license header at the top of the file applies.
# See docs/copyright_license.rst


def _cmark_encode_unknown(buf: _cmarkCmarkStrbuf):
    # static const uint8_t repl[] = {239, 191, 189};

    # man 3 printf
    #
    # #include <stdio.h>
    # #include <stdint.h>
    #
    # int main(void)
    # {
    #     static const uint8_t repl[] = {239, 191, 189};
    #     printf ("%hhu", repl);
    #
    #     return 0;
    # }
    #
    # gcc a.c
    # ./a.out
    # 9
    #
    # python3
    # >>> chr(9)
    # '\t'
    repl = '\t'
    _cmark_cmark_strbuf_put(buf, repl, 3)


# 0.29, 0.30
def _cmark_cmark_utf8proc_charlen(line: str, line_length: int) -> int:
    length: int
    i: int

    if not line_length:
        return 0

    # Use length = 1 instead of the utf8proc_utf8class[256]
    # list.
    # Python:
    #     length = utf8proc_utf8class[ord(line[0])]
    # C:
    #     length = utf8proc_utf8class[str[0]];
    # For example:
    # len('ł') == 2 # in Python 2
    # len('ł') == 1 # in Python 3
    # See the documentation.
    # In Python 3 all strings are unicode by default
    # and they all have length of 1.
    length = 1

    #     if (!length)
    #       return -1;

    if line_length >= 0 and length > line_length:
        return -line_length

    for i in range(1, length):
        if (ord(line[i]) & 0xC0) != 0x80:
            return -i

    return length


# 0.29, 0.30
def _cmark_cmark_utf8proc_iterate(line: str, line_len: int) -> tuple:
    length: int = 0
    uc: int = -1
    dst: int = -1

    length = _cmark_cmark_utf8proc_charlen(line, line_len)
    if length < 0:
        return -1, dst

    if length == 1:
        uc = ord(line[0])

    # In Python 3 all strings are unicode by default
    # and they all have length of 1.
    # All the original C code here is omitted for this reason.

    if uc < 0:
        return -1, dst

    dst = uc

    return length, dst


# 0.30
def _cmark_cmark_utf8proc_encode_char(uc: int, buf: _cmarkCmarkStrbuf):
    dst: str
    length: int = 0

    if uc < 0:
        raise ValueError

    # In Python 3 all strings are unicode by default
    # and they all have length of 1.
    # Omitted code.
    length = 1
    if uc > 1 or uc >= 0x110000:
        _cmark_encode_unknown(buf)
        return

    dst = str(uc)
    _cmark_cmark_strbuf_put(buf, dst, length)


# 0.30
def _cmark_cmark_utf8proc_case_fold(
    dest: _cmarkCmarkStrbuf,
    string: str,
    length: int,
):
    c: int

    while (length > 0):
        char_len, c = _cmark_cmark_utf8proc_iterate(string, length)

        if char_len >= 0:
            # FIXME: unsure about this. See original C source code.
            _cmark_cmark_utf8proc_encode_char(
                ord(unicodedata.normalize('NFC', chr(c)).casefold()), dest)
        else:
            _cmark_encode_unknown(dest)
            char_len = -char_len

        # Advance pointer.
        #     str += char_len;
        string = string[char_len:]
        # Reduce string length.
        length -= char_len


# 0.29, 0.30
def _cmark_cmark_utf8proc_is_space(char: int) -> bool:
    r"""Match anything in the Zs class, plus LF, CR, TAB, FF."""
    return (unicodedata.category(chr(char)) == 'Zs'
            or chr(char) in ['\u0009', '\u000A', '\u000C', '\u000D'])


# 0.29, 0.30
def _cmark_cmark_utf8proc_is_punctuation(char: int) -> bool:
    r"""Match anything in the P[cdefios] classes."""
    return ((char < 128 and _cmark_cmark_ispunct(char))
            or unicodedata.category(chr(char)).startswith('P'))


if __name__ == '__main__':
    pass



#
# scanners_h.py
#
# Copyright (C) 2017-2022 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
r"""A cmark implementation file."""

from .chunk_h import _cmarkCmarkChunk
from .scanners_c import _cmark__scan_at

# License C applies to this file except for non derivative code:
# in that case the license header at the top of the file applies.
# See docs/copyright_license.rst


def _cmark_scan_spacechars(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_spacechars', c, n)


def _cmark_scan_link_title(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_link_title', c, n)


def _cmark_scan_autolink_uri(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_autolink_uri', c, n)


def _cmark_scan_autolink_email(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_autolink_email', c, n)


def _cmark_scan_html_comment(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_html_comment', c, n)


def _cmark_scan_html_cdata(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_cdata', c, n)


def _cmark_scan_html_tag(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_html_tag', c, n)


def _cmark_scan_html_declaration(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_html_declaration', c, n)


def _cmark_scan_html_pi(c: _cmarkCmarkChunk, n: int) -> int:
    return _cmark__scan_at('_cmark__scan_html_pi', c, n)



#
# md-toc documentation build configuration file, created by
# sphinx-quickstart on Wed Dec 27 17:32:50 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
r"""conf.py."""

import os
import sys

sys.path.insert(0, os.path.abspath('..'))

# -- General configuration ------------------------------------------------

# General information about the project.
project = 'md-toc'
copyright = '2017-2024, Franco Masotti'
author = 'Franco Masotti'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '9.0.0'
# The full version, including alpha/beta/rc tags.
release = '9.0.0'

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.autosummary', 'sphinx.ext.coverage',
    'sphinx.ext.graphviz', 'sphinx_copybutton'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'default'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = ['css/custom.css']

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'md-toc-doc'

# -- Options for LaTeX output ---------------------------------------------

latex_engine = 'xelatex'
latex_elements = {

    # The paper size ('letterpaper' or 'a4paper').
    #
    'papersize': 'a4paper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        'md-toc.tex',
        'md-toc Documentation',
        'Franco Masotti',
        'manual',
    ),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        'md-toc',
        'md-toc Documentation',
        [author],
        1,
    ),
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'md-toc',
        'md-toc Documentation',
        author,
        'md-toc',
        'Generate table of contents for markdown files.',
        'Miscellaneous',
    ),
]

html_theme_options = {
    'repository_provider':
    'github',
    'repository_url':
    'https://software.franco.net.eu.org/frnmst/md-toc',
    'use_repository_button':
    True,
    'use_download_button':
    True,
    'use_issues_button':
    True,
    'announcement':
    '⚠️ starting from version 9 all the functions are only accessible via the full module path. For example: md_toc.build_toc(...) is now md_toc.api.build_toc(...) ⚠️'
}

html_baseurl = 'https://docs.franco.net.eu.org/md-toc/'

pygments_style = 'default'
html_last_updated_fmt = '%Y-%m-%d %H:%M:%S %z'
copybutton_line_continuation_character = '\\'

# Epub.
epub_theme = 'epub'
epub_author = 'Franco Masotti'
epub_theme_options = {
    'relbar1': False,
    'footer': False,
}
epub_css_style = [
    'css/epub.css',
]



#
# houdini_h.py
#
# Copyright (C) 2017-2022 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
r"""A cmark implementation file."""

# License F applies to this file except for non derivative code:
# in that case the license header at the top of the file applies.
# See docs/copyright_license.rst


def _cmark_HOUDINI_ESCAPED_SIZE(x: int) -> float:
    return (x * 12) / 10


def _cmark_HOUDINI_UNESCAPED_SIZE(x: int) -> int:
    return x



#
# __main__.py
#
# Copyright (C) 2017-2020 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
"""Call the CLI parser."""

import sys
import traceback

from .cli import CliInterface


def main(args=None):
    """Call the CLI interface and wait for the result."""
    retcode = 0
    try:
        ci = CliInterface()
        args = ci.parser.parse_args()
        result = args.func(args)
        if result is not None and not isinstance(result, bool):
            print(result)
        retcode = 0

        # TOC differs.
        if result:
            retcode = 128

    except Exception:
        retcode = 1
        traceback.print_exc()
    sys.exit(retcode)


if __name__ == '__main__':
    main()



#
# cmark_h.py
#
# Copyright (C) 2017-2022 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
r"""A cmark implementation file."""

from enum import Enum

# License C applies to this file except for non derivative code:
# in that case the license header at the top of the file applies.
# See docs/copyright_license.rst


# Defines the memory allocation functions to be used by CMark
# when parsing and allocating a document tree
#     typedef struct cmark_mem {
#       void *(*calloc)(size_t, size_t);
#       void *(*realloc)(void *, size_t);
#       void (*free)(void *);
#     } cmark_mem;
class _cmarkCmarkMem:
    pass


# C enum
# typedef enum { ... } cmark_node_type;
# Undefined value in the C source code get their value
# accoring to their position in the sequence, like an array.
class _cmarkCmarkNodeType(Enum):
    # Error status.
    CMARK_NODE_NONE = 0

    # Block.
    CMARK_NODE_DOCUMENT = 1
    CMARK_NODE_BLOCK_QUOTE = 2
    CMARK_NODE_LIST = 3
    CMARK_NODE_ITEM = 4
    CMARK_NODE_CODE_BLOCK = 5
    CMARK_NODE_HTML_BLOCK = 6
    CMARK_NODE_CUSTOM_BLOCK = 7
    CMARK_NODE_PARAGRAPH = 8
    CMARK_NODE_HEADING = 9
    CMARK_NODE_THEMATIC_BREAK = 10

    CMARK_NODE_FIRST_BLOCK = CMARK_NODE_DOCUMENT
    CMARK_NODE_LAST_BLOCK = CMARK_NODE_THEMATIC_BREAK

    # Inline
    CMARK_NODE_TEXT = 11
    CMARK_NODE_SOFTBREAK = 12
    CMARK_NODE_LINEBREAK = 13
    CMARK_NODE_CODE = 14
    CMARK_NODE_HTML_INLINE = 15
    CMARK_NODE_CUSTOM_INLINE = 16
    CMARK_NODE_EMPH = 17
    CMARK_NODE_STRONG = 18
    CMARK_NODE_LINK = 19
    CMARK_NODE_IMAGE = 20

    CMARK_NODE_FIRST_INLINE = CMARK_NODE_TEXT
    CMARK_NODE_LAST_INLINE = CMARK_NODE_IMAGE

    # For backwards compatibility:
    CMARK_NODE_HEADER = CMARK_NODE_HEADING
    CMARK_NODE_HRULE = CMARK_NODE_THEMATIC_BREAK
    CMARK_NODE_HTML = CMARK_NODE_HTML_BLOCK
    CMARK_NODE_INLINE_HTML = CMARK_NODE_HTML_INLINE


CMARK_OPT_SOURCEPOS = 1 << 1
CMARK_OPT_SMART = 1 << 10

if __name__ == '__main__':
    pass



#
# cli.py
#
# Copyright (C) 2017-2024 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
"""Command line interface file."""

import argparse
import sys
import textwrap
from importlib import metadata

from . import generic
from .api import (
    build_multiple_tocs,
    tocs_equal,
    write_strings_on_files_between_markers,
)
from .constants import common_defaults
from .constants import parser as md_parser

PROGRAM_DESCRIPTION = 'Markdown Table Of Contents: Automatically generate a compliant table\nof contents for a markdown file to improve document readability.'
VERSION_NAME = 'md_toc'
VERSION_COPYRIGHT = 'Copyright (C) 2017-2023 Franco Masotti, frnmst'
VERSION_LICENSE = 'License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\nThis is free software: you are free to change and redistribute it.\nThere is NO WARRANTY, to the extent permitted by law.'
RETURN_VALUES = 'Return values: 0 ok, 1 error, 2 invalid command, 128 TOC differs from the one in the file (see --diff option)'

try:
    VERSION_NUMBER = metadata.distribution('md_toc').version
except metadata.PackageNotFoundError:
    VERSION_NUMBER = 'development version (venv)'

PROGRAM_EPILOG = RETURN_VALUES + '\n\n' + VERSION_COPYRIGHT + '\n' + VERSION_LICENSE


class CliToApi():
    """An interface between the CLI and API functions."""

    def write_toc(self, args) -> bool:
        """Write the table of contents."""
        ordered = False
        if args.ordered_list_marker is not None:
            list_marker = args.ordered_list_marker
            ordered = True
        elif args.unordered_list_marker is not None:
            list_marker = args.unordered_list_marker
        newline_string = args.newline_string
        if newline_string == r'\n':
            newline_string = '\n'
        elif newline_string == r'\r':
            newline_string = '\r'
        if newline_string == r'\r\n':
            newline_string = '\r\n'

        toc_struct = build_multiple_tocs(
            filenames=args.filename,
            ordered=ordered,
            no_links=args.no_links,
            no_indentation=args.no_indentation,
            no_list_coherence=args.no_list_coherence,
            keep_header_levels=args.header_levels,
            parser=args.parser,
            list_marker=list_marker,
            skip_lines=args.skip_lines,
            constant_ordered_list=args.constant_ordered_list,
            newline_string=newline_string,
        )

        equal: bool = True

        if args.in_place:
            equal = write_strings_on_files_between_markers(
                filenames=args.filename,
                strings=toc_struct,
                marker=args.toc_marker,
                newline_string=newline_string,
            )
        else:
            for i, toc in enumerate(toc_struct):
                print(toc, end='')
                if args.diff:
                    equal &= tocs_equal(toc, args.filename[i], args.toc_marker)

        diff: bool = False
        if equal is False:
            diff = True

        # Return that the TOC(s) is differing only if the `--diff` argument
        # was selected, i.e.: `args.diff`.
        return diff & args.diff


class CliInterface():
    """The interface exposed to the final user."""

    def __init__(self):
        """Set the parser variable that will be used instead of using create_parser."""
        self.parser = self.create_parser()

    def _add_filename_argument(self, parser):
        # The filename argument is common to all markdown parsers.
        # See commit b65cf32.
        parser.add_argument(
            'filename',
            metavar='FILE_NAME',
            nargs='*',
            help='the I/O file name',
        )

    def _add_cmark_like_megroup_arguments(self, megroup, parser_name: str):
        megroup.add_argument(
            '-u',
            '--unordered-list-marker',
            choices=md_parser[parser_name]['list']['unordered']
            ['bullet_markers'],
            nargs='?',
            const=md_parser[parser_name]['list']['unordered']
            ['default_marker'],
            default=md_parser[parser_name]['list']['unordered']
            ['default_marker'],
            help='set the marker and enables unordered list. Defaults to ' +
            md_parser[parser_name]['list']['unordered']['default_marker'],
        )
        megroup.add_argument(
            '-o',
            '--ordered-list-marker',
            choices=md_parser[parser_name]['list']['ordered']
            ['closing_markers'],
            nargs='?',
            const=md_parser[parser_name]['list']['ordered']
            ['default_closing_marker'],
            help=('set the marker and enable ordered lists. Defaults to ' +
                  md_parser[parser_name]['list']['ordered']
                  ['default_closing_marker']),
        )

    def _add_cmark_like_arguments(self, parser, parser_name: str):
        parser.add_argument(
            '-c',
            '--constant-ordered-list',
            action='store_true',
            help='intead of progressive numbers use a single integer as \
                  list marker. This options enables ordered lists',
        )
        parser.add_argument(
            '-l',
            '--header-levels',
            type=int,
            choices=range(1,
                          md_parser[parser_name]['header']['max_levels'] + 1),
            nargs='?',
            const=md_parser[parser_name]['header']['default_keep_levels'],
            help='set the maximum level of headers to be considered as part \
                  of the TOC. Defaults to ' +
            str(md_parser[parser_name]['header']['default_keep_levels'], ),
        )

    def create_parser(self):
        """Create the CLI parser."""
        parser = argparse.ArgumentParser(
            description=PROGRAM_DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=textwrap.dedent(PROGRAM_EPILOG),
        )

        # markdown parser is first positional argument.
        # File names are arguments to the subparser.
        subparsers = parser.add_subparsers(
            dest='parser',
            title='markdown parser',
            help='<markdown parser> --help',
        )
        subparsers.required = True

        #########################
        # github + commonmarker #
        #########################
        github = subparsers.add_parser(
            'github',
            aliases=['commonmarker'],
            description='Use GitHub Flavored Markdown rules to generate \
                         an output. If no \
                         option is selected, the default output will be an \
                         unordered list with the respective default values \
                         as listed below',
        )
        self._add_filename_argument(github)
        megroup = github.add_mutually_exclusive_group()
        self._add_cmark_like_megroup_arguments(megroup, 'github')
        self._add_cmark_like_arguments(github, 'github')
        github.set_defaults(header_levels=md_parser['github']['header']
                            ['default_keep_levels'], )

        ##########
        # gitlab #
        ##########
        gitlab = subparsers.add_parser(
            'gitlab',
            description='Use GitLab Flavored Markdown rules to generate an \
                         output. If no \
                         option is selected, the default output will be an \
                         unordered list with the respective default values \
                         as listed below',
        )
        self._add_filename_argument(gitlab)
        megroup = gitlab.add_mutually_exclusive_group()
        self._add_cmark_like_megroup_arguments(megroup, 'gitlab')
        self._add_cmark_like_arguments(gitlab, 'gitlab')
        gitlab.set_defaults(header_levels=md_parser['gitlab']['header']
                            ['default_keep_levels'], )

        ####################
        # cmark + Goldmark #
        ####################
        cmark = subparsers.add_parser(
            'cmark',
            aliases=['goldmark'],
            description='Use CommonMark rules to generate an output. If no \
                         option is selected, the default output will be an \
                         unordered list with the respective default values \
                         as listed below',
        )
        self._add_filename_argument(cmark)
        megroup = cmark.add_mutually_exclusive_group()
        self._add_cmark_like_megroup_arguments(megroup, 'cmark')
        self._add_cmark_like_arguments(cmark, 'cmark')
        cmark.set_defaults(header_levels=md_parser['cmark']['header']
                           ['default_keep_levels'], )

        #############
        # Redcarpet #
        #############
        redcarpet = subparsers.add_parser(
            'redcarpet',
            description='Use Redcarpet rules to generate an output. If no \
                         option is selected, the default output will be an \
                         unordered list with the respective default values \
                         as listed below. Gitlab rules are the same as \
                         Redcarpet except that conflicts are avoided with \
                         duplicate headers.',
        )
        self._add_filename_argument(redcarpet)
        megroup = redcarpet.add_mutually_exclusive_group()
        megroup.add_argument(
            '-u',
            '--unordered-list-marker',
            choices=md_parser['redcarpet']['list']['unordered']
            ['bullet_markers'],
            nargs='?',
            const=md_parser['redcarpet']['list']['unordered']
            ['default_marker'],
            default=md_parser['redcarpet']['list']['unordered']
            ['default_marker'],
            help='set the marker and enables unordered list. Defaults to ' +
            md_parser['redcarpet']['list']['unordered']['default_marker'],
        )
        megroup.add_argument(
            '-o',
            '--ordered-list-marker',
            choices=md_parser['redcarpet']['list']['ordered']
            ['closing_markers'],
            nargs='?',
            const=md_parser['redcarpet']['list']['ordered']
            ['default_closing_marker'],
            help='set the marker and enables ordered lists. Defaults to ' +
            md_parser['redcarpet']['list']['ordered']
            ['default_closing_marker'],
        )
        redcarpet.add_argument(
            '-c',
            '--constant-ordered-list',
            action='store_true',
            help='intead of progressive numbers use a single integer as list \
                  marker. This options enables ordered lists',
        )
        redcarpet.add_argument(
            '-l',
            '--header-levels',
            type=int,
            choices=range(
                1,
                md_parser['redcarpet']['header']['max_levels'] + 1,
            ),
            nargs='?',
            const=md_parser['redcarpet']['header']['default_keep_levels'],
            help='set the maximum level of headers to be considered as part \
                  of the TOC. Defaults to ' +
            str(md_parser['redcarpet']['header']['default_keep_levels'], ),
        )
        redcarpet.set_defaults(header_levels=md_parser['redcarpet']['header']
                               ['default_keep_levels'], )

        ##########
        # Common #
        ##########
        no_list_coherence_or_no_indentation = parser.add_mutually_exclusive_group(
        )
        no_list_coherence_or_no_indentation.add_argument(
            '-c',
            '--no-list-coherence',
            action='store_true',
            help='avoids checking for TOC list coherence',
        )
        no_list_coherence_or_no_indentation.add_argument(
            '-i',
            '--no-indentation',
            action='store_true',
            help='avoids adding indentations to the TOC',
        )

        parser.add_argument(
            '-d',
            '--diff',
            action='store_true',
            help=(
                'returns 128 if the newly generated TOC differs from the one \
                  already existing in the file'),
        )
        parser.add_argument(
            '-l',
            '--no-links',
            action='store_true',
            help='avoids adding links to the TOC',
        )
        parser.add_argument(
            '-m',
            '--toc-marker',
            metavar='TOC_MARKER',
            default=common_defaults['toc_marker'],
            help=(
                'set the string to be used as the marker for positioning the \
                table of contents. Put this value between single quotes. \
                Defaults to \'' + common_defaults['toc_marker']) + '\'',
        )
        parser.add_argument(
            '-n',
            '--newline-string',
            choices=[r'\n', r'\r', r'\r\n'],
            metavar='NEWLINE_STRING',
            type=str,
            default=common_defaults['newline_string'],
            help=('the string used to separate the lines of the TOC. \
                  Use single quotes to delimit the string. \
                  If you output in place all the newlines of the \
                  input file will be replaced with this value. \
                  Defaults to ' + repr(common_defaults['newline_string'])),
        )
        parser.add_argument(
            '-p',
            '--in-place',
            action='store_true',
            help='overwrite the input file',
        )
        parser.add_argument(
            '-s',
            '--skip-lines',
            metavar='SKIP_LINES',
            type=int,
            default=0,
            help='skip parsing of the first selected number of lines. \
                  Defaults to 0, i.e. do not skip any lines',
        )
        parser.add_argument(
            '-v',
            '--version',
            action='version',
            version=VERSION_NAME + ' ' + VERSION_NUMBER,
        )

        parser.set_defaults(func=CliToApi().write_toc)

        return parser



#
# api.py
#
# Copyright (C) 2017-2024 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
"""The main file."""

from __future__ import annotations

import copy
import hashlib
import os
import re
import sys
import typing

import fpyutils

from . import generic, types
from .cmark import inlines_c, node_h, references_h
from .constants import common_defaults
from .constants import parser as md_parser
from .exceptions import (
    GithubEmptyLinkLabel,
    GithubOverflowCharsLinkLabel,
    GithubOverflowOrderedListMarker,
    StdinIsNotAFileToBeWritten,
    StringCannotContainNewlines,
    TocDoesNotRenderAsCoherentList,
)


def tocs_equal(current_toc: str, filename: str, marker: str) -> bool:
    r"""Check if the TOC already present in a file is the samw of the one passed to this function.

    :parameter current_toc: the new or current TOC. Do not include the ``<!--TOC-->\n\n`` and ``\n\n<!--TOC-->``.
    :parameter filename: the filename with the TOC for the comparison already present in the file.
    :parameter marker: the TOC marker.
    :type current_toc: str
    :type filename: str
    :type marker: str
    :returns: ``True`` if the two TOCs are the same, ``False`` otherwise
    :rtype: bool
    :raises: a built-in exception.
    """
    tocs_equal: bool = False
    r: tuple = generic._get_existing_toc(filename, marker)
    old_toc: str = r[0]
    if current_toc.strip() == old_toc:
        tocs_equal = True
    return tocs_equal


def write_string_on_file_between_markers(
    filename: str,
    string: str,
    marker: str,
    newline_string: str = common_defaults['newline_string'],
) -> bool:
    r"""Write the table of contents on a single file.

    :parameter filename: the file that needs to be read or modified.
    :parameter string: the string that will be written on the file.
    :parameter marker: a marker that will identify the start
         and the end of the string.
    :parameter newline_string: the new line separator.
         Defaults to ``os.linesep``.
    :type filename: str
    :type string: str
    :type marker: str
    :type newline_string: str
    :returns: ``True`` if new TOC is the same as the exising one, ``False`` otherwise.
    :rtype: bool
    :raises: StdinIsNotAFileToBeWritten or an fpyutils exception
         or a built-in exception.
    """
    if filename == '-':
        raise StdinIsNotAFileToBeWritten

    # TOC that is written to the file.
    final_toc_string: str = ''.join([
        marker, newline_string, newline_string,
        string.rstrip(), newline_string, newline_string, marker, newline_string
    ])

    (
        old_toc,
        lines_to_delete,
        two_or_more_markers,
        marker_line_positions_length,
        marker_line_positions,
        first_marker_line_number,
    ) = generic._get_existing_toc(filename, marker)

    equal: bool = False
    if string.strip() == old_toc:
        equal = True

    generic._remove_line_intervals(filename, lines_to_delete)

    # Only 1 pre-existing marker.
    if not two_or_more_markers and marker_line_positions_length == 1:
        generic._remove_line_intervals(
            filename, [[marker_line_positions[1], marker_line_positions[1]]])

    # 2 or more pre-existing markers.
    if marker_line_positions_length >= 1:
        fpyutils.filelines.insert_string_at_line(
            filename,
            final_toc_string,
            first_marker_line_number,
            filename,
            append=False,
            newline_character=newline_string,
        )

    return equal


def write_strings_on_files_between_markers(
    filenames: list[str],
    strings: list[str],
    marker: str,
    newline_string: str = common_defaults['newline_string'],
) -> bool:
    r"""Write the table of contents on multiple files.

    :parameter filenames: the files that needs to be read or modified.
    :parameter strings: the strings that will be written on the file. Each
         string is associated with one file.
    :parameter marker: a marker that will identify the start
         and the end of the string.
    :parameter newline_string: the new line separator.
         Defaults to ``os.linesep``.
    :type filenames: list
    :type strings: list
    :type marker: str
    :type newline_string: str
    :returns: ``True`` if all TOCs are the same as the existing ones, ``False``
         otherwise.
    :rtype: bool
    :raises: an fpyutils exception or a built-in exception.
    """
    if not len(filenames) == len(strings):
        raise ValueError

    file_id: int = 0
    equal: bool = True
    for f in filenames:
        equal &= write_string_on_file_between_markers(f, strings[file_id],
                                                      marker, newline_string)
        file_id += 1

    return equal


def build_toc(
    filename: str,
    ordered: bool = False,
    no_links: bool = False,
    no_indentation: bool = False,
    no_list_coherence: bool = False,
    keep_header_levels: int = 3,
    parser: str = 'github',
    list_marker: str = '-',
    skip_lines: int = 0,
    constant_ordered_list: bool = False,
    newline_string: str = common_defaults['newline_string'],
) -> str:
    r"""Build the table of contents of a single file.

    :parameter filename: the file that needs to be read.
    :parameter ordered: decides whether to build an ordered list or not.
        Defaults to ``False``.
    :parameter no_links: disables the use of links.
        Defaults to ``False``.
    :parameter no_indentation: disables indentation in the list.
        Defaults to ``False``.
    :parameter no_list_coherence: if set to ``False`` checks
        header levels for consecutiveness. If they are not consecutive
        an exception is raised. For example: ``# ONE\n### TWO\n``
        are not consecutive header levels while ``# ONE\n## TWO\n``
        are. Defaults to ``False``.
    :parameter keep_header_levels: the maximum level of headers to be
        considered as such when building the table of contents.
        Defaults to ``3``.
    :parameter parser: decides rules on how to generate anchor links.
        Defaults to ``github``.
    :parameter list_marker: a string that contains some of the first
        characters of the list element.
        Defaults to ``-``.
    :parameter skip_lines: the number of lines to be skipped from
        the start of file before parsing for table of contents.
        Defaults to ``0```.
    :parameter constant_ordered_list: use a single integer
        as list marker. This sets ordered to ``True``.
    :parameter newline_string: the newline separator.
        Defaults to ``os.linesep``.
    :type filename: str
    :type ordered: bool
    :type no_links: bool
    :type no_indentation: bool
    :type no_list_coherence: bool
    :type keep_header_levels: int
    :type parser: str
    :type list_marker: str
    :type skip_lines: int
    :type constant_ordered_list: bool
    :type newline_string: str
    :returns: toc, the corresponding table of contents of the file.
    :rtype: str
    :raises: a built-in exception.

    .. warning:: In case of ordered TOCs you must explicitly pass one of the
        supported ordered list markers.

    :Example:

    >>> import md_toc # doctest: +SKIP
    >>> with open('foo.md', 'w') as f: # doctest: +SKIP
    ...     f.write('# This\n# Is an\n## Example\n') # doctest: +SKIP
    26
    >>> md_toc.api.build_toc('foo.md') # doctest: +SKIP
    - [This](#this)
    - [Is an](#is-an)
      - [Example](#example)
    """
    if not skip_lines >= 0:
        raise ValueError

    toc: list[str] = []
    header_type_counter: types.HeaderTypeCounter = {}
    header_type_curr: int = 0
    header_type_prev: int = 0
    header_type_first: int = 0
    header_duplicate_counter: types.HeaderDuplicateCounter = {}

    if filename == '-':
        f = sys.stdin
    else:
        # When reading input from the stream,
        # if newline is None, universal newlines mode is enabled.
        # Lines in the input can end in '\n', '\r', or '\r\n',
        # and these are translated into '\n' before being returned to the caller.
        # See
        # https://docs.python.org/3/library/functions.html#open-newline-parameter
        f = open(filename, newline=None)

    # Help the developers: override the list_marker in case
    # this function is called with the default unordered list marker,
    # for example like this:
    # print(md_toc.build_toc('test.md', ordered=True))
    # This avoids an AssertionError later on.
    if (ordered and list_marker
            == md_parser[parser]['list']['unordered']['default_marker']):
        list_marker = md_parser[parser]['list']['ordered'][
            'default_closing_marker']
    if constant_ordered_list:
        ordered = True
    if ordered and (
            list_marker is None or list_marker
            not in md_parser[parser]['list']['ordered']['closing_markers']):
        list_marker = md_parser[parser]['list']['ordered'][
            'default_closing_marker']

    if skip_lines > 0:
        loop: bool = True
        line_counter = 1
        while loop:
            if line_counter > skip_lines or f.readline() == '':
                loop = False
            line_counter += 1

    try:
        line = f.readline()
    except UnicodeDecodeError:
        return ''.join(
            ['<!--stop reading ', filename, ': probably a binary file-->'])

    indentation_log: dict[types.IndentationLogElement] = init_indentation_log(
        parser, list_marker)
    is_within_code_fence = False
    code_fence = None
    is_document_end = False
    while line:
        # Document ending detection.
        #
        # This changes the state of is_within_code_fence if the
        # file has no closing fence markers. This serves no practial
        # purpose since the code would run correctly anyway. It is
        # however more semantically correct.
        #
        # See the unit tests (examples 95 and 96 of the github parser)
        # and the is_closing_code_fence function.
        if filename != '-':
            # stdin is not seekable.
            file_pointer_pos = f.tell()
            try:
                if f.readline() == '':
                    is_document_end = True
            except UnicodeDecodeError:
                return ''.join([
                    '<!--stop reading ', filename,
                    ': probably a binary file-->'
                ])
            f.seek(file_pointer_pos)

        # Code fence detection.
        if is_within_code_fence:
            is_within_code_fence = not is_closing_code_fence(
                line,
                code_fence,
                is_document_end,
                parser,
            )
        else:
            code_fence = is_opening_code_fence(line, parser)
            if code_fence is not None:
                # Update the status of the next line.
                is_within_code_fence = True

        if not is_within_code_fence or code_fence is None:

            # Header detection and gathering.
            headers: list[types.Header] = get_md_header(
                line,
                header_duplicate_counter,
                keep_header_levels,
                parser,
                no_links,
            )

            # We only need to get the first element since all the lines
            # have been already separated here.
            header: types.Header = headers[0]

            # Ignore invalid or to-be-invisible headers.
            if header is not None and header['visible']:
                header_type_curr = header['header_type']

                # Take care of the ordered TOC.
                if ordered and not constant_ordered_list:
                    increase_index_ordered_list(
                        header_type_counter,
                        header_type_prev,
                        header_type_curr,
                        parser,
                    )
                    index = header_type_counter['h' + str(header_type_curr)]
                elif constant_ordered_list:
                    # This value should work on most parsers.
                    index = 1
                else:
                    index = 1

                # Take care of list indentations.
                if no_indentation:
                    no_of_indentation_spaces_curr = 0
                    # TOC list coherence checks are not necessary
                    # without indentation.
                else:
                    if not no_list_coherence:
                        # In-place list coherence checks can be made using only
                        # the first, current and previous header types.
                        if header_type_first == 0:
                            header_type_first = header_type_curr
                        if header_type_prev == 0:
                            header_type_prev = header_type_curr
                        if (header_type_curr < header_type_first
                                or header_type_curr > header_type_prev + 1):
                            raise TocDoesNotRenderAsCoherentList

                    compute_toc_line_indentation_spaces(
                        header_type_curr,
                        header_type_prev,
                        parser,
                        ordered,
                        list_marker,
                        indentation_log,
                        index,
                    )
                    no_of_indentation_spaces_curr = indentation_log[
                        header_type_curr]['indentation_spaces']

                # endif

                # Build a single TOC line.
                toc_line_no_indent = build_toc_line_without_indentation(
                    header,
                    ordered,
                    no_links,
                    index,
                    parser,
                    list_marker,
                )

                # Save the TOC line with the indentation.
                toc.append(''.join([
                    build_toc_line(
                        toc_line_no_indent,
                        no_of_indentation_spaces_curr,
                    ), newline_string
                ]))

                header_type_prev = header_type_curr

            # endif

        # endif

        try:
            line = f.readline()
        except UnicodeDecodeError:
            return ''.join(
                ['<!--stop reading ', filename, ': probably a binary file-->'])

    # endwhile

    f.close()

    return ''.join(toc)


def build_multiple_tocs(
    filenames: list[str],
    ordered: bool = False,
    no_links: bool = False,
    no_indentation: bool = False,
    no_list_coherence: bool = False,
    keep_header_levels: int = 3,
    parser: str = 'github',
    list_marker: str = '-',
    skip_lines: int = 0,
    constant_ordered_list: bool = False,
    newline_string: str = common_defaults['newline_string'],
) -> list[str]:
    r"""Parse files by line and build the table of contents of each file.

    :parameter filenames: the files that needs to be read.
    :parameter ordered: decides whether to build an ordered list or not.
         Defaults to ``False``.
    :parameter no_links: disables the use of links.
         Defaults to ``False``.
    :parameter no_indentation: disables indentation in the list.
         Defaults to ``False``.
    :parameter keep_header_levels: the maximum level of headers to be
         considered as such when building the table of contents.
         Defaults to ``3``.
    :parameter parser: decides rules on how to generate anchor links.
         Defaults to ``github``.
    :parameter skip_lines: the number of lines to be skipped from
         the start of file before parsing for table of contents.
         Defaults to ``0```.
    :parameter list_marker: a string that contains some of the first
         characters of the list element.
         Defaults to ``-``.
    :parameter constant_ordered_list: use a single integer
        as list marker. This sets ordered to ``True``.
    :parameter newline_string: the newline separator.
        Defaults to ``os.linesep``.
    :type filenames: list
    :type ordered: bool
    :type no_links: bool
    :type no_indentation: bool
    :type keep_header_levels: int
    :type parser: str
    :type list_marker: str
    :type skip_lines: int
    :type constant_ordered_list: bool
    :type newline_string: str
    :returns: toc_struct, the corresponding table of contents for each input
         file.
    :rtype: list[str]
    :raises: a built-in exception.

    .. warning:: In case of ordered TOCs you must explicitly pass one of the
        supported ordered list markers.
    """
    if len(filenames) == 0:
        filenames.append('-')
    return [
        build_toc(
            filenames[file_id],
            ordered,
            no_links,
            no_indentation,
            no_list_coherence,
            keep_header_levels,
            parser,
            list_marker,
            skip_lines,
            constant_ordered_list,
            newline_string,
        ) for file_id in range(0, len(filenames))
    ]


def increase_index_ordered_list(
    header_type_count: types.HeaderTypeCounter,
    header_type_prev: int,
    header_type_curr: int,
    parser: str = 'github',
):
    r"""Compute the current index for ordered list table of contents.

    :parameter header_type_count: the count of each header type.
    :parameter header_type_prev: the previous type of header (h[1,...,Inf]).
    :parameter header_type_curr: the current type of header (h[1,...,Inf]).
    :parameter parser: decides rules on how to generate ordered list markers.
         Defaults to ``github``.
    :type header_type_count: types.HeaderTypeCounter
    :type header_type_prev: int
    :type header_type_curr: int
    :type parser: str
    :returns: None
    :rtype: None
    :raises: GithubOverflowOrderedListMarker or a built-in exception.
    """
    # header_type_prev might be 0 while header_type_curr can't.
    if not header_type_prev >= 0:
        raise ValueError
    if not header_type_curr >= 1:
        raise ValueError

    # Base cases for a new table of contents or a new index type.
    if header_type_prev == 0:
        header_type_prev = header_type_curr
    if ('h' + str(header_type_curr) not in header_type_count
            or header_type_prev < header_type_curr):
        header_type_count['h' + str(header_type_curr)] = 0

    header_type_count['h' + str(header_type_curr)] += 1

    if parser in ['github', 'cmark', 'gitlab', 'commonmarker']:
        if header_type_count['h' + str(header_type_curr)] > md_parser[
                'github']['list']['ordered']['max_marker_number']:
            raise GithubOverflowOrderedListMarker


def init_indentation_log(
    parser: str = 'github',
    list_marker: str = '-',
) -> dict[types.IndentationLogElement]:
    r"""Create a data structure that holds list marker information.

    :parameter parser: decides rules on how compute indentations.
         Defaults to ``github``.
    :parameter list_marker: a string that contains some of the first
         characters of the list element. Defaults to ``-``.
    :type parser: str
    :type list_marker: str
    :returns: indentation_log, the data structure.
    :rtype: dict[types.IndentationLogElement]
    :raises: a built-in exception.

    .. warning: this function does not make distinctions between ordered and unordered
         TOCs.
    """
    return {
        i: {
            'index': md_parser[parser]['list']['ordered']['min_marker_number'],
            'list_marker': list_marker,
            'indentation_spaces': 0,
        }
        for i in range(1, md_parser[parser]['header']['max_levels'] + 1)
    }


def compute_toc_line_indentation_spaces(
    header_type_curr: int = 1,
    header_type_prev: int = 0,
    parser: str = 'github',
    ordered: bool = False,
    list_marker: str = '-',
    indentation_log: dict[types.IndentationLogElement] = init_indentation_log(
        'github', '-'),
    index: int = 1,
):
    r"""Compute the number of indentation spaces for the TOC list element.

    :parameter header_type_curr: the current type of header (h[1,...,Inf]).
         Defaults to ``1``.
    :parameter header_type_prev: the previous type of header (h[1,...,Inf]).
         Defaults to ``0``.
    :parameter parser: decides rules on how compute indentations.
         Defaults to ``github``.
    :parameter ordered: if set to ``True``, numbers will be used
         as list ids instead of dash characters.
         Defaults to ``False``.
    :parameter list_marker: a string that contains some of the first
         characters of the list element.
         Defaults to ``-``.
    :parameter indentation_log: a data structure that holds list marker
         information for ordered lists.
         Defaults to ``init_indentation_log('github', '.')``.
    :parameter index: a number that will be used as list id in case of an
         ordered table of contents. Defaults to ``1``.
    :type header_type_curr: int
    :type header_type_prev: int
    :type parser: str
    :type ordered: bool
    :type list_marker: str
    :type indentation_log: dict[types.IndentationLogElement]
    :type index: int
    :returns: None
    :rtype: None
    :raises: a built-in exception.

    .. warning:: In case of ordered TOCs you must explicitly pass one of the
        supported ordered list markers.
    """
    if not header_type_curr >= 1:
        raise ValueError
    if not header_type_prev >= 0:
        raise ValueError
    if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:
        if not len(indentation_log
                   ) == md_parser['github']['header']['max_levels']:
            raise ValueError
    if not index >= 1:
        raise ValueError

    if parser in [
            'github', 'cmark', 'gitlab', 'commonmarker', 'goldmark',
            'redcarpet'
    ]:
        if ordered:
            if list_marker not in md_parser[parser]['list']['ordered'][
                    'closing_markers']:
                raise ValueError
        else:
            if list_marker not in md_parser[parser]['list']['unordered'][
                    'bullet_markers']:
                raise ValueError

    if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:
        index_length: int
        if ordered:
            if header_type_prev == 0:
                index_length = 0
            else:
                index_length = len(
                    str(indentation_log[header_type_prev]['index']), )
        else:
            index_length = 0

        if header_type_prev == 0:
            # Base case for the first toc line.
            indentation_log[header_type_curr]['indentation_spaces'] = 0
        elif header_type_curr > header_type_prev:
            # More indentation.
            indentation_log[header_type_curr]['indentation_spaces'] = (
                indentation_log[header_type_prev]['indentation_spaces'] +
                len(indentation_log[header_type_prev]['list_marker'], ) +
                index_length + len(' '))
        elif header_type_curr < header_type_prev:
            # Less indentation. Since we went "back" we must reset
            # all the sublists in the data structure. The indentation spaces are the ones
            # computed before.
            for i in range(
                    header_type_curr + 1,
                    md_parser['github']['header']['max_levels'] + 1,
            ):
                indentation_log[i]['index'] = md_parser['github']['list'][
                    'ordered']['min_marker_number']
                indentation_log[i]['indentation_spaces'] = 0
                indentation_log[i]['list_marker'] = list_marker
        # And finally, in case of same indentation we have: header_type_curr = header_type_prev
        # so indentation_log[header_type_curr]['indentation_spaces'] = indentation_log[header_type_prev]['indentation_spaces']
        # which is an identity.

        if ordered:
            indentation_log[header_type_curr]['index'] = index
        indentation_log[header_type_curr]['list_marker'] = list_marker

    elif parser in ['redcarpet']:
        indentation_log[header_type_curr]['indentation_spaces'] = 4 * (
            header_type_curr - 1)


def build_toc_line_without_indentation(
    header: types.Header,
    ordered: bool = False,
    no_links: bool = False,
    index: int = 1,
    parser: str = 'github',
    list_marker: str = '-',
) -> str:
    r"""Return a list element of the table of contents.

    :parameter header: a data structure that contains the original
         text, the trimmed text and the type of header.
    :parameter ordered: if set to ``True``, numbers will be used
         as list ids, otherwise a dash character. Defaults
         to ``False``.
    :parameter no_links: disables the use of links. Defaults to ``False``.
    :parameter index: a number that will be used as list id in case of an
         ordered table of contents. Defaults to ``1``.
    :parameter parser: decides rules on how compute indentations.
         Defaults to ``github``.
    :parameter list_marker: a string that contains some of the first
         characters of the list element. Defaults to ``-``.
    :type header: types.Header
    :type ordered: bool
    :type no_links: bool
    :type index: int
    :type parser: str
    :type list_marker: str
    :returns: toc_line_no_indent, a single line of the table of contents
         without indentation.
    :rtype: str
    :raises: a built-in exception.

    .. warning:: In case of ordered TOCs you must explicitly pass one of the
        supported ordered list markers.
    """
    if not header['header_type'] >= 1:
        raise ValueError
    if not index >= 1:
        raise ValueError
    if parser in [
            'github', 'cmark', 'gitlab', 'commonmarker', 'goldmark',
            'redcarpet'
    ]:
        if ordered:
            if list_marker not in md_parser[parser]['list']['ordered'][
                    'closing_markers']:
                raise ValueError
        else:
            if list_marker not in md_parser[parser]['list']['unordered'][
                    'bullet_markers']:
                raise ValueError

    if parser in [
            'github', 'cmark', 'gitlab', 'commonmarker', 'goldmark',
            'redcarpet'
    ]:
        if ordered:
            list_marker: str = ''.join([str(index), list_marker])

        # See
        # https://spec.commonmark.org/0.28/#example-472
        # and
        # https://github.com/vmg/redcarpet/blob/e3a1d0b00a77fa4e2d3c37322bea66b82085486f/ext/redcarpet/markdown.c#L998
        line: str
        if no_links:
            line = header['text_original']
        else:
            line = ''.join([
                '[', header['text_original'], ']', '(#',
                header['text_anchor_link'], ')'
            ])

        return ''.join([list_marker, ' ', line])


def build_toc_line(
    toc_line_no_indent: str,
    no_of_indentation_spaces: int = 0,
) -> str:
    r"""Build the TOC line.

    :parameter toc_line_no_indent: the TOC line without indentation.
    :parameter no_of_indentation_spaces: the number of indentation spaces.
         Defaults to ``0``.
    :type toc_line_no_indent: str
    :type no_of_indentation_spaces: int
    :returns: toc_line, a single line of the table of contents.
    :rtype: str
    :raises: a built-in exception.

    :Example:

    >>> import md_toc
    >>> md_toc.api.build_toc_line('', 0)
    ''

    :Example:

    >>> import md_toc
    >>> md_toc.api.build_toc_line('my string', 10)
    '          my string'
    """
    if not no_of_indentation_spaces >= 0:
        raise ValueError

    indentation: str = no_of_indentation_spaces * ' '
    return ''.join([indentation, toc_line_no_indent])


def remove_html_tags(line: str, parser: str = 'github') -> str:
    r"""Remove HTML tags.

    :parameter line: a string.
    :parameter parser: decides rules on how to remove HTML tags.
         Defaults to ``github``.
    :type line: str
    :type parser: str
    :returns: the input string without HTML tags.
    :rtype: str
    :raises: a built-in exception.
    """
    if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:
        # We need to match newline as well because it is a WS, so we
        # must use re.DOTALL.
        regexes: list[str] = [
            md_parser[parser]['re']['OT'], md_parser[parser]['re']['CT'],
            md_parser[parser]['re']['CO'], md_parser[parser]['re']['PI'],
            md_parser[parser]['re']['DE'], md_parser[parser]['re']['CD']
        ]

        for r in regexes:
            line = re.sub(r, '', line, flags=re.DOTALL)

    return line


def remove_emphasis(line: str, parser: str = 'github') -> str:
    r"""Remove markdown emphasis.

    :parameter line: a string.
    :parameter parser: decides rules on how to find delimiters.
        Defaults to ``github``.
    :type line: str
    :type parser: str
    :returns: the input line without emphasis.
    :rtype: str
    :raises: a built-in exception.

    .. note:: Backslashes are preserved.

    :Example:

    >>> import md_toc
    >>> md_toc.api.remove_emphasis('__my string__ *is this* one')
    'my string is this one'
    """
    if parser in [
            'github', 'cmark', 'gitlab', 'commonmarker', 'goldmark',
            'redcarpet'
    ]:
        mem = None
        refmap = references_h._cmarkCmarkReferenceMap()

        parent = node_h._cmarkCmarkNode()

        # Remove NULL bytes.
        line = line.replace('\x00', '')

        parent.data = line
        parent.length = len(line)
        parent.start_line = 0
        parent.start_column = 0
        parent.internal_offset = 1

        ignore: list[range] = inlines_c._cmark_cmark_parse_inlines(
            mem, parent, refmap, 0)
        line = filter_indices_from_line(line, ignore)

    elif parser in ['redcarpet']:
        # TODO
        pass

    return line


def filter_indices_from_line(line: str, ranges: list[range]) -> str:
    r"""Given a line and a Python ranges, remove the characters in the ranges.

    :parameter line: a string.
    :parameter ranges: a list of Python ranges.
    :type line: str
    :type ranges: list
    :returns: the line without the specified indices.
    :rtype: str
    :raises: a built-in exception.
    """
    # Transform ranges into lists.
    rng: list[typing.Sequence[int]] = [list(r) for r in ranges]

    # Flatten list.
    rng_flat: list[int] = sorted([item for e in rng for item in e])

    return ''.join([line[i] for i in range(0, len(line)) if i not in rng_flat])


def anchor_link_punctuation_filter(
    input_string: str,
    parser: str = 'github',
) -> str:
    r"""Remove punctuation and other unwanted characters from the anchor link string.

    :parameter input_string: the unfiltered anchor link
    :parameter parser: decides rules on how to generate anchor links.
        Defaults to ``github``.
    :type input_string: str
    :type parser: str
    :returns: a string without the unwanted characters.
    :rtype: str
    :raises: a built-in exception.

    .. note: license B applies for the github part.
        See docs/copyright_license.rst
    """
    # https://github.com/gjtorikian/html-pipeline/blob/7c7fad1f82f81ebf15dd81d59eed28d979b8e441/lib/html/pipeline/toc_filter.rb#L30
    output_string: str
    if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:
        # Remove punctuation: Keep spaces, hypens and "word characters" only.
        # In other words, given the set:
        # \w                alphanumeric
        # \-                hyphen
        #                   space
        # remove its complementary set from the input_string. This is achieved
        # using `^`.
        pattern: str = r'[^\w\- ]'
        replacement: str = ''
        output_string = re.sub(pattern, replacement, input_string)
    return output_string


def build_anchor_link(
    header_text_trimmed: str,
    header_duplicate_counter: types.HeaderDuplicateCounter,
    parser: str = 'github',
) -> str:
    r"""Apply the specified slug rule to build the anchor link.

    :parameter header_text_trimmed: the text that needs to be transformed
        in a link.
    :parameter header_duplicate_counter: a data structure that keeps track of
        possible duplicate header links in order to avoid them. This is
        meaningful only for certain values of parser.
    :parameter parser: decides rules on how to generate anchor links.
        Defaults to ``github``.
    :type header_text_trimmed: str
    :type header_duplicate_counter: types.HeaderDuplicateCounter
    :type parser: str
    :returns: None if the specified parser is not recognized, or the anchor
        link, otherwise.
    :rtype: str
    :raises: a built-in exception.

    .. note: license A applies for the redcarpet part.
        See docs/copyright_license.rst

    :Example:

    >>> import md_toc
    >>> md_toc.api.build_anchor_link('This is an example test       header', {}, 'gitlab')
    'this-is-an-example-test-header'
    """
    # Check for newlines.
    if len(replace_and_split_newlines(header_text_trimmed)) > 1:
        raise StringCannotContainNewlines

    if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:
        header_text_trimmed = header_text_trimmed.lower()

        # Filter HTML tags.
        header_text_trimmed = remove_html_tags(header_text_trimmed, parser)

        # Filter "emphasis and strong emphasis".
        header_text_trimmed = remove_emphasis(header_text_trimmed, parser)

        # Remove punctuation characters.
        header_text_trimmed = anchor_link_punctuation_filter(
            header_text_trimmed, parser)

        # Replace spaces with dash.
        header_text_trimmed = header_text_trimmed.replace(' ', '-')

        if parser in ['gitlab']:
            # See https://docs.gitlab.com/ee/user/markdown.html#header-ids-and-links
            # Two or more hyphens in a row are converted to one.
            header_text_trimmed = re.sub('-+', '-', header_text_trimmed)

        # Use a checksum to reduce memory usage on the dict:
        #
        # import sys
        # size = getsizeof(header_duplicate_counter)
        # size += sum(map(getsizeof, header_duplicate_counter.values())) + sum(map(getsizeof, header_duplicate_counter.keys()))
        #
        # See also
        # https://stackoverflow.com/a/6579816
        ht_checksum = hashlib.sha1(
            header_text_trimmed.encode('UTF-8')).hexdigest()
        # Check for duplicates.
        # Set the initial value if we are examining the first occurrency.
        # The state of header_duplicate_counter is available to the caller
        # functions.
        if ht_checksum not in header_duplicate_counter:
            header_duplicate_counter[ht_checksum] = 0
        if header_duplicate_counter[ht_checksum] > 0:
            header_text_trimmed = ''.join([
                header_text_trimmed, '-',
                str(header_duplicate_counter[ht_checksum])
            ])
        header_duplicate_counter[ht_checksum] += 1
        return header_text_trimmed
    elif parser in ['redcarpet']:
        # To ensure full compatibility what follows is a direct translation
        # of the rndr_header_anchor C function used in redcarpet.
        STRIPPED = " -&+$,/:;=?@\"#{}|^~[]`\\*()%.!'"
        header_text_trimmed_len = len(header_text_trimmed)
        inserted = 0
        stripped = 0
        header_text_trimmed_middle_stage = list()
        for i in range(0, header_text_trimmed_len):
            if header_text_trimmed[i] == '<':
                while i < header_text_trimmed_len and header_text_trimmed[
                        i] != '>':
                    i += 1
            elif header_text_trimmed[i] == '&':
                while i < header_text_trimmed_len and header_text_trimmed[
                        i] != ';':
                    i += 1
            # str.find() == -1 if character is not found in str.
            # https://docs.python.org/3.6/library/stdtypes.html?highlight=find#str.find
            elif not generic._isascii(header_text_trimmed[i]) or STRIPPED.find(
                    header_text_trimmed[i], ) != -1:
                if inserted and not stripped:
                    header_text_trimmed_middle_stage.append('-')
                stripped = 1
            else:
                header_text_trimmed_middle_stage.append(
                    header_text_trimmed[i].lower(), )
                stripped = 0
                inserted += 1

        header_text_trimmed_middle_stage_str: str = ''.join(
            header_text_trimmed_middle_stage)
        if stripped > 0 and inserted > 0:
            header_text_trimmed_middle_stage_str = header_text_trimmed_middle_stage_str[
                0:-1]

        if inserted == 0 and header_text_trimmed_len > 0:
            hash: int = 5381
            for i in range(0, header_text_trimmed_len):
                # Get the unicode representation with ord.
                # Unicode should be equal to ASCII in ASCII's range of
                # characters.
                hash = ((hash << 5) + hash) + ord(header_text_trimmed[i])

            # This is equivalent to %x in C. In Python we don't have
            # the length problem so %x is equal to %lx in this case.
            # Apparently there is no %l in Python...
            header_text_trimmed_middle_stage_str = 'part-' + '{:x}'.format(
                hash)

        return header_text_trimmed_middle_stage_str

    return None


def replace_and_split_newlines(line: str) -> list[str]:
    r"""Replace all the newline characters with line feeds and separate the components.

    :parameter line: a string.
    :type line: str
    :returns: a list of newline separated strings.
    :rtype: list[str]
    :raises: a built-in exception.
    """
    line = line.replace('\r\n', '\n')
    line = line.replace('\r', '\n')

    # Ignore the sequences of consecutive first and last newline characters.
    # Since there is no need to process the last
    # empty line.
    line = line.lstrip('\n')
    line = line.rstrip('\n')

    return line.split('\n')


def get_atx_heading(
    line: str,
    keep_header_levels: int = 3,
    parser: str = 'github',
    no_links: bool = False,
) -> list[types.AtxHeadingStructElement]:
    r"""Given a line extract the link label and its type.

    :parameter line: the line to be examined. This string may include newline
        characters in between.
    :parameter keep_header_levels: the maximum level of headers to be
         considered as such when building the table of contents.
         Defaults to ``3``.
    :parameter parser: decides rules on how to generate the anchor text.
         Defaults to ``github``.
    :parameter no_links: disables the use of links.
    :type line: str
    :type keep_header_levels: int
    :type parser: str
    :type no_links: bool
    :returns: struct, a list of dictionaries
    :rtype: list[types.AtxHeadingStructElement]
    :raises: GithubEmptyLinkLabel or GithubOverflowCharsLinkLabel or a
         built-in exception.

    .. note: license A applies for the redcarpet part.
         See docs/copyright_license.rst

    .. note:: license B applies for the github part.
         See docs/copyright_license.rst
    """
    if not keep_header_levels >= 1:
        raise ValueError

    struct: list[types.AtxHeadingStructElement] = list()
    line_visible: bool = True

    if len(line) == 0:
        return [{
            'header_type': None,
            'header_text_trimmed': None,
            'visible': False
        }]

    for subl in replace_and_split_newlines(line):
        current_headers = None
        final_line = None

        if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:

            struct.append({
                'header_type': None,
                'header_text_trimmed': None,
                'visible': False
            })

            # Empty substring or backslash.
            if len(subl) == 0 or subl[0] == '\u005c':
                continue

            # Preceding.
            i: int = 0
            while i < len(subl) and subl[i] == ' ' and i <= md_parser[
                    'github']['header']['max_space_indentation']:
                i += 1
            if i > md_parser['github']['header']['max_space_indentation']:
                continue

            # ATX characters.
            offset = i
            while i < len(subl) and subl[i] == '#' and i <= md_parser[
                    'github']['header']['max_levels'] + offset:
                i += 1

            if (i - offset > md_parser['github']['header']['max_levels']
                    or i - offset == 0):
                continue

            # We need to continue parsing to find possible duplicate headers
            # in other functions if we set keep_header_levels < max_levels.
            if i - offset > keep_header_levels:
                line_visible: bool = False

            current_headers = i - offset

            # At this moment GFM is still at version 0.29
            # while cmark is at 0.30. There are subtle differences
            # such as this one. Assume gitlab is on par with 0.30.
            if parser in ['github', 'commonmarker']:
                # GFM 0.29 and cmark 0.29.
                spaces = [' ']
            else:
                # cmark 0.30.
                spaces = [' ', '\u0009']

            # Include special cases for line endings which should not be
            # discarded as non-ATX headers.
            if i < len(subl) and subl[i] not in spaces + ['\u000a', '\u000d']:
                continue

            i += 1

            # Exclude leading whitespaces after the ATX header identifier.
            # [0.29]:
            #   The opening sequence of `#` characters must be followed
            #   by a space or by the end of line.
            # [0.30]:
            #   The opening sequence of `#` characters must be followed
            #   by spaces or tabs, or by the end of line.
            while i < len(subl) and subl[i] in spaces:
                i += 1

            # An algorithm to find the start and the end of the closing sequence (cs).
            # The closing sequence includes all the significant part of the
            # string. This algorithm has a complexity of O(n) with n being the
            # length of the line.
            cs_start = i
            cs_end = cs_start
            # subl_prime =~ subl'.
            subl_prime = subl[::-1]
            len_subl = len(subl)
            hash_char_rounds = 0
            go = True
            i = 0
            hash_round_start = i

            # Ignore all characters after newlines and carrage returns which
            # are not at the end of the line.
            # See the two CRLF marker tests.
            crlf_marker = 0
            stripped_crlf = False
            while i < len_subl - cs_start:
                if subl_prime[i] in ['\u000a', '\u000d']:
                    crlf_marker = i
                    stripped_crlf = True
                i += 1

            # crlf_marker is the first CR LF character in the string.
            i = crlf_marker
            if stripped_crlf:
                # Skip last character only if is '\u000a', '\u000d'.
                i += 1

            # We know for sure that from now '\u000a', '\u000d' will not be
            # considered.
            cs_end = i

            # Cut spaces and hashes.
            while go and i < len_subl - cs_start:
                if (subl_prime[i] not in spaces + ['#']
                        or hash_char_rounds > 1):
                    if i > hash_round_start and hash_char_rounds > 0:
                        cs_end = len_subl - hash_round_start
                    else:
                        cs_end = len_subl - i
                    go = False
                if go:
                    while subl_prime[i] in spaces:
                        i += 1
                    hash_round_start = i
                    while subl_prime[i] == '#':
                        i += 1
                    if i > hash_round_start:
                        hash_char_rounds += 1

            final_line = subl[cs_start:cs_end]

            # Add escaping.
            if not no_links:
                if len(final_line) > 0 and final_line[-1] == '\u005c':
                    final_line = ''.join([final_line, ' '])
                if len(
                        final_line.strip('\u0020').strip('\u0009').strip(
                            '\u000a').strip('\u000b').strip('\u000c').strip(
                                '\u000d'), ) == 0:
                    raise GithubEmptyLinkLabel
                if len(final_line,
                       ) > md_parser['github']['link']['max_chars_label']:
                    raise GithubOverflowCharsLinkLabel

                i = 0
                while i < len(final_line):
                    # Escape square brackets if not already escaped.
                    if (final_line[i] == '[' or final_line[i] == ']'):
                        j = i - 1
                        consecutive_escape_characters = 0
                        while j >= 0 and final_line[j] == '\u005c':
                            consecutive_escape_characters += 1
                            j -= 1
                        if ((consecutive_escape_characters > 0
                             and consecutive_escape_characters % 2 == 0)
                                or consecutive_escape_characters == 0):
                            tmp = '\u005c'
                        else:
                            tmp = ''
                        final_line = ''.join(
                            [final_line[0:i], tmp, final_line[i:]])
                        i += 1 + len(tmp)
                    else:
                        i += 1

            # Overwrite the element with None as values.
            struct[-1] = {
                'header_type': current_headers,
                'header_text_trimmed': final_line,
                'visible': line_visible
            }

        elif parser in ['redcarpet']:
            struct.append({
                'header_type': None,
                'header_text_trimmed': None,
                'visible': False
            })

            if len(subl) == 0 or subl[0] != '#':
                continue

            i = 0
            while (i < len(subl)
                   and i < md_parser['redcarpet']['header']['max_levels']
                   and subl[i] == '#'):
                i += 1
            current_headers = i

            if i < len(subl) and subl[i] != ' ':
                continue

            while i < len(subl) and subl[i] == ' ':
                i += 1

            end = i
            while end < len(subl) and subl[end] != '\n':
                end += 1

            while end > 0 and subl[end - 1] == '#':
                end -= 1

            while end > 0 and subl[end - 1] == ' ':
                end -= 1

            if end > i:
                final_line = subl
                if not no_links and len(
                        final_line) > 0 and final_line[-1] == '\\':
                    final_line += ' '
                    end += 1
                final_line = final_line[i:end]

            if final_line is None:
                current_headers = None
                line_visible = False

            struct[-1] = {
                'header_type': current_headers,
                'header_text_trimmed': final_line,
                'visible': line_visible
            }

        # TODO: escape or remove '[', ']', '(', ')' in inline links for redcarpet,
        # TODO: check link label rules for redcarpet.
        # TODO: check live_visible

    # endfor

    return struct


def get_md_header(
    header_text_line: str,
    header_duplicate_counter: types.HeaderDuplicateCounter,
    keep_header_levels: int = 3,
    parser: str = 'github',
    no_links: bool = False,
) -> list[types.Header]:
    r"""Build a data structure with the elements needed to create a TOC line.

    :parameter header_text_line: a single markdown line that needs to be
        transformed into a TOC line. This line may include nmultiple newline
        characters in between.
    :parameter header_duplicate_counter: a data structure that contains the
         number of occurrencies of each header anchor link. This is used to
         avoid duplicate anchor links and it is meaningful only for certain
         values of parser.
    :parameter keep_header_levels: the maximum level of headers to be
         considered as such when building the table of contents.
         Defaults to ``3``.
    :parameter parser: decides rules on how to generate anchor links.
         Defaults to ``github``.
    :type header_text_line: str
    :type header_duplicate_counter: types.HeaderDuplicateCounter
    :type keep_header_levels: int
    :type parser: str
    :returns: a list with elements ``None`` if the input line does not correspond
        to one of the designated cases or a list of data structures containing
        the necessary components to create a table of contents.
    :rtype: list
    :raises: a built-in exception.

    .. note::
         This works like a wrapper to other functions.
    """
    return [
        None if
        (r['header_type'] is None and r['header_text_trimmed'] is None) else {
            'header_type':
            r['header_type'],
            'text_original':
            r['header_text_trimmed'],
            'text_anchor_link':
            build_anchor_link(
                r['header_text_trimmed'],
                header_duplicate_counter,
                parser,
            ),
            'visible':
            r['visible'],
        } for r in get_atx_heading(
            header_text_line,
            keep_header_levels,
            parser,
            no_links,
        )
    ]


def is_valid_code_fence_indent(line: str, parser: str = 'github') -> bool:
    r"""Determine if the given line has valid indentation for a code block fence.

    :parameter line: a single markdown line to evaluate.
    :parameter parser: decides rules on how to generate the anchor text.
         Defaults to ``github``.
    :type line: str
    :type parser: str
    :returns: True if the given line has valid indentation or False
         otherwise.
    :rtype: bool
    :raises: a built-in exception.
    """
    if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:
        return len(line) - len(line.lstrip(
            ' ')) <= md_parser['github']['code_fence']['min_marker_characters']
    elif parser in ['redcarpet']:
        # TODO.
        return False

    return False


def is_opening_code_fence(line: str, parser: str = 'github') -> str | None:
    r"""Determine if the given line is possibly the opening of a fenced code block.

    :parameter line: a single markdown line to evaluate.
    :parameter parser: decides rules on how to generate the anchor text.
         Defaults to ``github``.
    :type line: str
    :type parser: str
    :returns: None if the input line is not an opening code fence. Otherwise,
         returns the string which will identify the closing code fence
         according to the input parsers' rules.
    :rtype: typing.Optional[str]
    :raises: a built-in exception.
    """
    if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:
        markers = md_parser['github']['code_fence']['marker']
        marker_min_length = md_parser['github']['code_fence'][
            'min_marker_characters']

        info_string: str
        info_string_start: int

        if not is_valid_code_fence_indent(line):
            return None

        line = line.lstrip(' ').rstrip('\n')
        if not line.startswith((markers['backtick'] * marker_min_length,
                                markers['tilde'] * marker_min_length)):
            return None

        # The info string is the whole line except the first opening code
        # fence markers.
        if line == len(line) * line[0]:
            info_string = ''
            info_string_start = len(line)
        else:
            # Get the index where the info string starts.
            i: int = 0
            done: bool = False
            while not done and i < len(line):
                if line[i] != line[0]:
                    info_string_start = i
                    done = True
                i += 1

            info_string = line[info_string_start:len(line)]
            # [0.29]
            #   The line with the opening code fence may optionally contain
            #   some text following the code fence; this is trimmed of leading
            #   and trailing whitespace and called the info string.
            # [0.30]
            #   The line with the opening code fence may optionally contain
            #   some text following the code fence; this is trimmed of leading
            #   and trailing spaces or tabs and called the info string.
            # This does not change the end result of this function.
            if parser in ['github', 'commonmarker']:
                spaces = ' '
            else:
                spaces = ' \u0009'
            info_string = info_string.strip(spaces)

        # Backticks or tildes in info strings are explicitly forbidden
        # in a backtick-opened code fence, for Commonmark 0.29.
        # This also solves example 107 [Commonmark 0.28]. See:
        # https://github.github.com/gfm/#example-107
        if markers['backtick'] in info_string and line[0] != markers['tilde']:
            return None

        return line[0:info_string_start]
    elif parser in ['redcarpet']:
        # TODO.
        return None

    return None


def is_closing_code_fence(
    line: str,
    fence: str,
    is_document_end: bool = False,
    parser: str = 'github',
) -> bool:
    r"""Determine if the given line is the end of a fenced code block.

    :parameter line: a single markdown line to evaluate.
    :paramter fence: a sequence of backticks or tildes marking the start of
         the current code block. This is usually the return value of the
         is_opening_code_fence function.
    :parameter is_document_end: This variable tells the function that the
         end of the file is reached.
         Defaults to ``False``.
    :parameter parser: decides rules on how to generate the anchor text.
         Defaults to ``github``.
    :type line: str
    :type fence: str
    :type is_document_end: bool
    :type parser: str
    :returns: True if the line ends the current code block. False otherwise.
    :rtype: bool
    :raises: a built-in exception.

    :Example:

    >>> import md_toc
    >>> md_toc.api.is_closing_code_fence('```', '```')
    True

    :Example:

    >>> import md_toc
    >>> md_toc.api.is_closing_code_fence('```', '~~~')
    False
    """
    if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark']:
        markers = md_parser['github']['code_fence']['marker']
        marker_min_length = md_parser['github']['code_fence'][
            'min_marker_characters']

        if not is_valid_code_fence_indent(line):
            return False

        # Remove opening fence indentation after it is known to be valid.
        fence = fence.lstrip(' ')
        # Check if fence uses valid characters.
        if not fence.startswith((markers['backtick'], markers['tilde'])):
            return False

        if len(fence) < marker_min_length:
            return False

        # Additional security.
        fence = fence.rstrip(' \n')
        # Check that all fence characters are equal.
        if fence != len(fence) * fence[0]:
            return False

        # We might be inside a code block if this is not closed
        # by the end of the document, according to example 96 and 97.
        # This means that the end of the document corresponds to
        # a closing code fence.
        # Of course we first have to check that fence is a valid opening
        # code fence marker.
        # See:
        # example 96 [Commonmark 0.29]
        # example 97 [Commonmark 0.29]
        if is_document_end:
            return True

        # Check if line uses the same character as fence.
        line = line.lstrip(' ')
        if not line.startswith(fence):
            return False

        # [0.29]
        #   The closing code fence may be indented up to three spaces, and may be
        #   followed only by spaces, which are ignored.
        # [0.30]
        #   The closing code fence may be preceded by up to three spaces of
        #   indentation, and may be followed only by spaces or tabs,
        #   which are ignored.
        if parser in ['github', 'commonmarker']:
            spaces = ' '
        else:
            spaces = ' \u0009'
        line = line.rstrip(spaces + '\n')

        # Solves:
        # example 93 -> 94 [Commonmark 0.28]
        # example 94 -> 95 [Commonmark 0.29]
        if len(line) < len(fence):
            return False

        # Closing fence must not have alien characters.
        if line != len(line) * line[0]:
            return False

        return True
    elif parser in ['redcarpet']:
        # TODO.
        return False

    return False


if __name__ == '__main__':
    pass



#
# scanners_c.py
#
# Copyright (C) 2017-2022 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
r"""A cmark implementation file."""

import copy
import re

from ..constants import parser as md_parser
from .chunk_h import _cmarkCmarkChunk

# License C applies to this file except for non derivative code:
# in that case the license header at the top of the file applies.
# See docs/copyright_license.rst

# These functions have been re-written to avoid GOTO jumps,
# also using the scanners.re source file.
# The original C source states:
# /* Generated by re2c 1.3 */


def _cmark__scan_at(
    scanner_function_name: str,
    c: _cmarkCmarkChunk,
    offset: int,
) -> int:
    res: int
    ptr: str = c.data

    if ptr is None or offset > c.length:
        return 0
    else:
        #     lim: str = ptr[c.length]

        #     ptr[c.length] = '\0'
        if scanner_function_name == '_cmark__scan_spacechars':
            res = _cmark__scan_spacechars(ptr, offset)
        if scanner_function_name == '_cmark__scan_link_title':
            res = _cmark__scan_link_title(ptr, offset)
        if scanner_function_name == '_cmark__scan_autolink_uri':
            res = _cmark__scan_autolink_uri(ptr, offset)
        if scanner_function_name == '_cmark__scan_autolink_email':
            res = _cmark__scan_autolink_email(ptr, offset)
        if scanner_function_name == '_cmark__scan_html_comment':
            res = _cmark__scan_html_comment(ptr, offset)
        if scanner_function_name == '_cmark__scan_cdata':
            res = _cmark__scan_cdata(ptr, offset)
        if scanner_function_name == '_cmark__scan_html_tag':
            res = _cmark__scan_html_tag(ptr, offset)
        if scanner_function_name == '_cmark__scan_html_declaration':
            res = _cmark__scan_html_declaration(ptr, offset)
        if scanner_function_name == '_cmark__scan_html_pi':
            res = _cmark__scan_html_pi(ptr, offset)

        #     ptr[c.length] = lim

    return res


def _common_scan(regex: str, ptr: str, p: int) -> int:
    start_match: int = 0
    end_match: int = 0
    retval: int

    span = re.match(regex, ptr[p:])
    if span:
        ll = list(span.span())
        start_match = ll[0]
        end_match = ll[1]
        retval = end_match - start_match
    else:
        retval = 0

    return retval


# Try to match a link title (in single quotes, in double quotes, or
# in parentheses), returning number of chars matched.  Allow one
# level of internal nesting (quotes within quotes).
def _cmark__scan_link_title(ptr: str, p: int) -> int:
    r1 = '["](' + md_parser['cmark']['_scanners.re'][
        'escaped_char'] + '|[^"\u0000])*["]'
    r2 = "['](" + md_parser['cmark']['_scanners.re'][
        'escaped_char'] + "|[^'\u0000])*[']"
    r3 = r'[\(](' + md_parser['cmark']['_scanners.re'][
        'escaped_char'] + r"|[^\(\)\u0000])*[']"
    r = '(' + r1 + '|' + r2 + '|' + r3 + ')'
    return _common_scan(r, ptr, p)


# Match SOME space characters, including newlines.
def _cmark__scan_spacechars(ptr: str, p: int) -> int:
    return _common_scan(md_parser['cmark']['_scanners.re']['spacechar'] + '+',
                        ptr, p)


# Try to match URI autolink after first <, returning number of chars matched.
def _cmark__scan_autolink_uri(ptr: str, p: int) -> int:
    return _common_scan('[:][^\x00-\x20<>]*[>]', ptr, p)


# Try to match email autolink after first <, returning num of chars matched.
def _cmark__scan_autolink_email(ptr: str, p: int) -> int:
    return _common_scan(
        '[a-zA-Z0-9.!#$%&\'*+/=?^_`{|}~-]+[@][a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?([.][a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*[>]',
        ptr, p)


def _cmark__scan_html_comment(ptr: str, p: int) -> int:
    return _common_scan(md_parser['cmark']['_scanners.re']['htmlcomment'], ptr,
                        p)


def _cmark__scan_cdata(ptr: str, p: int) -> int:
    return _common_scan(md_parser['cmark']['_scanners.re']['cdata'], ptr, p)


# Try to match an HTML tag after first <, returning num of chars matched.
def _cmark__scan_html_tag(ptr: str, p: int) -> int:
    return _common_scan(md_parser['cmark']['_scanners.re']['htmltag'], ptr, p)


def _cmark__scan_html_declaration(ptr: str, p: int) -> int:
    return _common_scan(md_parser['cmark']['_scanners.re']['declaration'], ptr,
                        p)


def _cmark__scan_html_pi(ptr: str, p: int) -> int:
    return _common_scan(
        md_parser['cmark']['_scanners.re']['processinginstruction'], ptr, p)



#
# reference_c.py
#
# Copyright (C) 2017-2022 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
r"""A cmark implementation file."""

import functools

from ..constants import parser as md_parser
from .buffer_c import (
    _cmark_cmark_strbuf_detach,
    _cmark_cmark_strbuf_normalize_whitespace,
    _cmark_cmark_strbuf_trim,
)
from .buffer_h import _cmark_CMARK_BUF_INIT, _cmarkCmarkStrbuf
from .chunk_h import _cmarkCmarkChunk
from .cmark_h import _cmarkCmarkMem
from .references_h import _cmarkCmarkReference, _cmarkCmarkReferenceMap
from .utf8_c import _cmark_cmark_utf8proc_case_fold

# License C applies to this file except for non derivative code:
# in that case the license header at the top of the file applies.
# See docs/copyright_license.rst


# normalize reference:  collapse internal whitespace to single space,
# remove leading/trailing whitespace, case fold
# Return NULL if the reference name is actually empty (i.e. composed
# solely from whitespace)
# 0.30
def _cmark_normalize_reference(mem: _cmarkCmarkMem,
                               ref: _cmarkCmarkChunk) -> str:
    normalized: _cmarkCmarkStrbuf = _cmark_CMARK_BUF_INIT(mem)
    result: str

    if ref is None:
        return None

    if ref.length == 0:
        return None

    _cmark_cmark_utf8proc_case_fold(normalized, ref.data, ref.length)
    _cmark_cmark_strbuf_trim(normalized)
    _cmark_cmark_strbuf_normalize_whitespace(normalized)

    result = _cmark_cmark_strbuf_detach(normalized)
    if not result:
        raise ValueError

    #     if result[0] == '\0':
    if result == '':
        #     mem.free(result)
        del result
        return None

    return result


def _cmark_labelcmp(a: str, b: str) -> bool:
    return a == b


def _cmark_refcmp(p1: _cmarkCmarkReference, p2: _cmarkCmarkReference) -> int:
    res: bool = _cmark_labelcmp(p1.label, p2.label)

    if res:
        return res
    else:
        return p1.age - p2.age


def _cmark_sort_references(maps: _cmarkCmarkReferenceMap):
    i: int
    last: int = 0
    size: int = maps.size
    r: _cmarkCmarkReference = maps.refs

    #     **sorted = NULL;
    # A list of _cmarkCmarkReference
    srt: list = list()

    #     sorted = (cmark_reference **)maps->mem->calloc(size, sizeof(cmark_reference *));
    while (r):
        srt.append(r)
        r = r.next

    #     qsort(sorted, size, sizeof(cmark_reference *), refcmp);
    srt = sorted(srt, key=functools.cmp_to_key(_cmark_refcmp))

    for i in range(1, size):
        if _cmark_labelcmp(srt[i].label, srt[last].label) != 0:
            last += 1
            srt[last] = srt[i]

    maps.sorted = srt
    maps.size = last + 1


# Returns reference if refmaps contains a reference with matching
# label, otherwise NULL.
# 0.30
def _cmark_cmark_reference_lookup(
    maps: _cmarkCmarkReferenceMap,
    label: _cmarkCmarkChunk,
) -> _cmarkCmarkReference:
    # A list of _cmarkCmarkReference
    ref: list = None
    r: _cmarkCmarkReference = None
    norm: str

    #     MAX_LINK_LABEL_LENGTH
    if label.length < 1 or label.length > md_parser['cmark']['link'][
            'max_chars_label'] + 1:
        return None

    if maps is None or not maps.size:
        return None

    norm = _cmark_normalize_reference(maps.mem, label)
    if norm is None:
        return None

    if not maps.sorted:
        _cmark_sort_references(maps)

    # TODO
    #      ref = (cmark_reference **)bsearch(norm, maps->sorted, maps->size, sizeof(cmark_reference *),
    #                    refsearch);
    # FIXME

    #     maps->mem->free(norm);
    del norm

    if ref is not None:
        r = ref[0]
        # Check for expansion limit
        if maps.max_ref_size and r.size > maps.max_ref_size - maps.ref_size:
            return None
        maps.ref_size += r.size

    return r



#
# constants.py
#
# Copyright (C) 2017-2024 Franco Masotti (see /README.md)
#
# This file is part of md-toc.
#
# md-toc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# md-toc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with md-toc.  If not, see <http://www.gnu.org/licenses/>.
#
"""A file that contains all the global constants."""

import copy
import html
import os
import typing

# License C applies to the cmark entities part.
# See docs/copyright_license.rst
_entities5 = html.entities.html5
# remove keys without semicolons.  For some reason the list
# has duplicates of a few things, like auml, one with and one
# without a semicolon.
_ents = sorted([(k[:-1], _entities5[k].encode('utf-8'))
                for k in _entities5.keys() if k[-1] == ';'])

# Use a list instead of a class for simplicity.
# For this reason the cmark/entities.inc file is missing.
_entities: list = list()
for (ent, bs) in _ents:
    _entities.append({
        'entity': ent,
        'bytes': (' '.join(map(str, bs)) + ' 0').split(' ')
    })
    # Transform each entity into a list of integers from a list of strings.
    _entities[-1]['bytes'] = [int(n) for n in _entities[-1]['bytes']]

# Regular expressions related to scanners functions.
# See scanners.re and scanners.c files.
__cmark_spacechar = '([ \t\v\f\r\n])'
__cmark_tagname = '([A-Za-z][A-Za-z0-9-]*)'
__cmark_attributename = '([a-zA-Z_:][a-zA-Z0-9:._-]*)'
__cmark_unquotedvalue = "([^ \t\r\n\v\f\"'=<>`\x00]+)"
__cmark_singlequotedvalue = "(['][^'\x00]*['])"
__cmark_doublequotedvalue = '(["][^"\x00]*["])'
__cmark_attributevalue = '(' + __cmark_unquotedvalue + '|' + __cmark_singlequotedvalue + '|' + __cmark_doublequotedvalue + ')'
__cmark_attributevaluespec = __cmark_spacechar + '*[=]' + __cmark_spacechar + '*' + __cmark_attributevalue
__cmark_attribute = '(' + __cmark_spacechar + '+' + __cmark_attributename + __cmark_attributevaluespec + '?)'
__cmark_opentag = __cmark_tagname + __cmark_attribute + '*' + __cmark_spacechar + '*[/]?[>]'
__cmark_closetag = '[/]' + __cmark_tagname + __cmark_spacechar + '*[>]'

common_defaults: dict = {
    'toc_marker': '<!--TOC-->',
    'newline_string': os.linesep,
}

parser: dict = {
    'cmark': {
        'list': {
            'ordered': {
                'closing_markers': ['.', ')'],
                'default_marker_number': 1,
                'min_marker_number': 0,
                'max_marker_number': 999999999,
                'default_closing_marker': '.',
            },
            'unordered': {
                'bullet_markers': ['-', '+', '*'],
                'default_marker': '-',
            },
        },
        'link': {
            'max_chars_label': 999,
        },
        'header': {
            'max_space_indentation': 3,
            'max_levels': 6,
            'default_keep_levels': 6,
        },
        'code_fence': {
            'marker': {
                'backtick': '`',
                'tilde': '~',
            },
            'min_marker_characters': 3,
        },
        # Regular expressions related to entities functions.
        # See make_entities_inc.py and entities.inc files.
        're': {
            'ENTITIES': {
                'CMARK_ENTITY_MIN_LENGTH': 2,
                'CMARK_ENTITY_MAX_LENGTH': 32,
                'CMARK_NUM_ENTITIES': len(_entities),
                'entities': _entities,
            },
            # [0.30] only.
            'SPACETAB': '[\u0009\u0020]',
            # Line ending.
            'LE': '(\u000a|\u000d|\u000d\u000a)',

            # See https://spec.commonmark.org/0.28/#raw-html
            # 1. Open tag and 2. close tag.
            'DQAV': __cmark_doublequotedvalue,
            'SQAV': __cmark_singlequotedvalue,
            'UAV': __cmark_unquotedvalue,

            # 2.
            'AN': __cmark_attributename,
            'TN': __cmark_tagname,

            # 3. HTML comment.
            'COS': '<!--',
            'COT': '((?!>|->)(?:(?!--).))+(?!-).?',
            'COE': '-->',

            # 4. Processing instructions.
            'PIS': r'<\?',
            'PIB': r'(?:(?!\?>).)*',
            'PIE': r'\?>',

            # 5. Declarations.
            'DES': '<!',
            'DEN': '[A-Z]+',
            'DEB': '(?:(?!>).)+',
            'DEE': '>',

            # 6. CDATA
            # Section.
            'CDS': r'<!\[CDATA\[',
            # Body.
            'CDB': r'(?:(?!\]\]>).)+',
            # End.
            'CDE': r'\]\]>',

            # Attribute value.
            'AV': __cmark_attributevalue,

            # Attribute value specification.
            'AVS': __cmark_attributevaluespec,
        },
        '_scanners.re': {
            # FIXME
            # Some of these expressions are a duplicate of parser['cmark']['re'] dicts.
            'spacechar': __cmark_spacechar,
            'escaped_char': '([\\][!"#$%&\'()*+,./:;<=>?@[\\\\]^_`{|}~-])',
            'cdata': r'CDATA\[([^\]\x00]+|\][^\]\x00]|\]\][^>\x00])*',
            'htmltag': '(' + __cmark_opentag + '|' + __cmark_closetag + ')',
            'htmlcomment': '(--->|(-([-]?[^\x00>-])([-]?[^\x00-])*-->))',
            'declaration': '[A-Z]+' + __cmark_spacechar + '+' + '[^>\x00]*',
            'processinginstruction': '([^?>\x00]+|[?][^>\x00]|[>])+',
        },
    },
    'redcarpet': {
        'list': {
            'ordered': {
                # FIXME
                'min_marker_number': 0,
                'closing_markers': ['.'],
                'default_closing_marker': '.',
            },
            'unordered': {
                'bullet_markers': ['-', '+', '*'],
                'default_marker': '-',
            },
        },
        'header': {
            'max_space_indentation': 0,
            'max_levels': 6,
            'default_keep_levels': 6,
        },
    },
}

parser['cmark']['re'].update({
    # Attribute.
    # [0.30]
    #   An attribute consists of spaces, tabs, and up to one line ending,
    #   an attribute name, and an optional attribute value specification.
    # A newline is needed if spaces are not present in the first part.
    'AT':
    ('(' + parser['cmark']['re']['SPACETAB'] + '+' + '|' +
     parser['cmark']['re']['LE'] + '{1,1}' + ')' +
     parser['cmark']['re']['AN'] + '(' + parser['cmark']['re']['AVS'] + ')?'),
})

parser['cmark']['re'].update({
    # 1. Open tag.
    'OT':
    ('<' + parser['cmark']['re']['TN'] + '(' + parser['cmark']['re']['AT'] +
     ')*' + '(' + parser['cmark']['re']['SPACETAB'] + '*' + '|' +
     parser['cmark']['re']['LE'] + '?' + ')' + '(/)?' + '>'),
    # 2. Close tag.
    'CT': ('</' + parser['cmark']['re']['TN'] + '(' +
           parser['cmark']['re']['SPACETAB'] + '*' + '|' +
           parser['cmark']['re']['LE'] + '?' + ')' + '>'),

    # 3. HTML comment.
    'CO':
    parser['cmark']['re']['COS'] + parser['cmark']['re']['COT'] +
    parser['cmark']['re']['COE'],

    # 4. Processing instructions.
    'PI':
    parser['cmark']['re']['PIS'] + parser['cmark']['re']['PIB'] +
    parser['cmark']['re']['PIE'],

    # 5. Declarations.
    'DE':
    parser['cmark']['re']['DES'] + parser['cmark']['re']['DEN'] +
    parser['cmark']['re']['DEB'] + parser['cmark']['re']['DEE'],

    # 6. CDATA.
    'CD':
    parser['cmark']['re']['CDS'] + parser['cmark']['re']['CDB'] +
    parser['cmark']['re']['CDE'],
})

parser['github'] = copy.deepcopy(parser['cmark'])

# FIXME
# The following overrides must be removed once GFM is on par with cmark 0.30.
# FIXME
# Regular expressions.
# These refer to inline HTML.
parser['github']['re'].update({
    'UAV':
    "[^\u0020\"'=<>`]+",
    'WS':
    '(\u0020|\u0009|\u000a|\u000b|\u000c|\u000d)',
})

parser['github']['re'].update({
    'AV': ('(' + parser['github']['re']['UAV'] + '|' +
           parser['github']['re']['SQAV'] + '|' +
           parser['github']['re']['DQAV'] + ')'),
    'AVS': (parser['github']['re']['WS'] + '*' + '=' +
            parser['github']['re']['WS'] + '*' + parser['github']['re']['AV']),

    # Attribute.
    'AT': (parser['github']['re']['WS'] + '+' + parser['github']['re']['AN'] +
           '(' + parser['github']['re']['AVS'] + ')?'),

    # Remember: https://developmentality.wordpress.com/2011/09/22/python-gotcha-word-boundaries-in-regular-expressions/
    # Github Flavored Markdown Disallowed Raw HTML (specific to GFM and not to cmark')
    # See
    # https://github.github.com/gfm/#disallowed-raw-html-extension-
    # This RE are specific to GFM.
    'GDRH':
    r'''(\b[tT][iI][tT][lL][eE]\b|\b[tT][eE][xX][tT][aA][rR][eE][aA]\b|\b[sS][tT][yY][lL][eE]\b|\b[xX][mM][pP]\b|\b[iI][fF][rR][aA][mM][eE]\b|\b[nN][oO][eE][mM][bB][eE][dD]\b|\b[nN][oO][fF][rR][aA][mM][eE][sS]\b|\b[sS][cC][rR][iI][pP][tT]\b|\b[pP][lL][aA][iI][nN][tT][eE][xX][tT]\b)''',
    'DEW':
    parser['github']['re']['WS'] + '+',
})

parser['github']['re'].update({
    'TN': ('(?!' + parser['github']['re']['GDRH'] + ')' +
           parser['github']['re']['TN']),
})

parser['github']['re'].update({
    # 1. Open tag.
    'OT':
    ('<' + parser['github']['re']['TN'] + '(' + parser['github']['re']['AT'] +
     ')*' + '(' + parser['github']['re']['WS'] + ')*' + '(/)?' + '>'),
    # 2. Close tag.
    'CT': ('</' + parser['github']['re']['TN'] + parser['github']['re']['WS'] +
           '?' + '>'),
    # 5. Declarations.
    'DE': (parser['github']['re']['DES'] + parser['github']['re']['DEN'] +
           parser['github']['re']['DEW'] + parser['github']['re']['DEB'] +
           parser['github']['re']['DEE']),
})

del parser['github']['re']['SPACETAB']
del parser['github']['re']['LE']

##########################################

# Do not move these.
parser['gitlab'] = copy.deepcopy(parser['cmark'])
parser['goldmark'] = copy.deepcopy(parser['cmark'])
parser['commonmarker'] = copy.deepcopy(parser['github'])

if __name__ == '__main__':
    pass
